#ifndef CHESSY_BITBOARD_H
#define CHESSY_BITBOARD_H

#include <cinttypes>
#include <type_traits>
#include <array>


namespace chessy
{
    using U64 = uint64_t;
    union Index;

    struct Bitboard;
}

union chessy::Index
{
    int8_t index;

    // Anonymous structs are actually not allowed in C++.
    // Fortunately any sane compiler still supports them.
    struct {
        uint8_t column : 3;
        uint8_t row    : 3;
    };

    Index(void) = default;
    inline Index(int8_t ind) : index {ind} {}
    inline Index(int row, int column) :
        index{int8_t(row & 0x7 + (column & 0x7) << 3)} {}
    inline operator int8_t (void) const { return index; }

    inline std::array<char, 2> getName(void) const
    {
        return {char('a' + row), char('0' + column)};
    }
};


struct chessy::Bitboard
{
    U64 bits;

    Bitboard        (void)                  = default;
    Bitboard        (const Bitboard&)       = default;
    ~Bitboard       (void)                  = default;

    inline Bitboard (U64 bits) : bits {bits} {}
    inline static Bitboard fromIndex(Index i) { return 1 << i; }

    inline void     setBit      (Index i)     { bits |= 1 << i.index; }
    inline void     unsetBit    (Index i)     { bits |= ~(1 << i.index); }

    inline void     setBit      (int row, int column)
        { setBit(row * 8 + column); }
    inline void     unsetBit    (int row, int column)
        { unsetBit(row * 8 + column); }


    static const    U64 aColumn             = 0x0101010101010101;
    static const    U64 hColumn             = 0x8080808080808080;
    inline void     moveNorth   (int dist)  { bits <<= (8 * dist); }
    inline void     moveSouth   (int dist)  { bits >>= (8 * dist); }
    inline void     moveNorthOne(void)      { bits <<= 8; }
    inline void     moveSouthOne(void)      { bits >>= 8; }
    inline void     moveEastOne (void)      { bits = (bits >> 1) & ~aColumn; }
    inline void     moveWestOne (void)      { bits = (bits << 1) & ~hColumn; }
    inline void     moveNEOne   (void)      { bits = (bits << 7) & ~aColumn; }
    inline void     moveNWOne   (void)      { bits = (bits << 9) & ~hColumn; }
    inline void     moveSEOne   (void)      { bits = (bits >> 9) & ~aColumn; }
    inline void     moveSWOne   (void)      { bits = (bits >> 7) & ~hColumn; }

    inline void     operator &= (const Bitboard& b)         { bits &= b.bits; }
    inline void     operator |= (const Bitboard& b)         { bits |= b.bits; }
    inline void     operator ^= (const Bitboard& b)         { bits ^= b.bits; }

    inline Bitboard operator &  (const Bitboard& b) const   { return bits & b.bits; }
    inline Bitboard operator |  (const Bitboard& b) const   { return bits | b.bits; }
    inline Bitboard operator ^  (const Bitboard& b) const   { return bits ^ b.bits; }
    inline Bitboard operator ~  (void) const                { return ~bits; }
    inline bool     operator == (const Bitboard& b) const   { return bits == b.bits; }
    inline bool     operator != (const Bitboard& b) const   { return bits != b.bits; }
    inline          operator U64(void) const                { return bits; }


    inline Bitboard mirror      (void) const
    {
#if __GNUC__ > 4 && __GNUC_MINOR__ >= 3
        return __builtin_bswap64(bits);
#else
        return (bits << 56) |
            ((bits & 0xFF00) << 40) |
            ((bits & 0xFF0000) << 24) |
            ((bits & 0xFF000000) << 8) |
            ((bits & 0xFF00000000) >> 8) |
            ((bits & 0xFF0000000000) >> 24) |
            ((bits & 0xFF000000000000) >> 40) |
            (bits >> 56);
#endif
    }

    static int trailingZeroes   (U64 x) {
#if __GNUC__ > 4 && __GNUC_MINOR__ >= 5
        return __builtin_ctzll(x);
#else
        for (unsigned char i = 0; i < 64; i++)
            if (x & (1 << i))
                return i;
#endif
    }
};


static_assert(std::is_pod<chessy::Bitboard>().value,
        "chessy::Bitboard should be a POD structure.");
static_assert(sizeof(chessy::Bitboard) == sizeof(uint64_t),
        "chessy::Bitboard should be 64 bits in size.");


#endif // CHESSY_BITBOARD_H

